"""Ubiquiti mFi MPower device"""

from __future__ import annotations

from typing import Any

from .entities import MPowerSensor, MPowerSwitch
from .interface import MPowerLED, MPowerNetwork, MPowerInterface


class MPowerDevice:
    """mFi mPower device representation."""

    _data: dict
    _interface: MPowerInterface

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
    ) -> None:
        """Initialize the device."""
        self._data = {}
        self._interface = MPowerInterface(host, username, password)

    async def __aenter__(self) -> MPowerDevice:
        """Enter context manager scope."""
        await self.interface.connect()
        await self.refresh()
        return self

    async def __aexit__(self, *kwargs) -> None:
        """Leave context manager scope."""
        await self.interface.close()

    def __str__(self) -> str:
        """Represent this device as string."""
        if self.has_data:
            keys = [
                "name",
                "model",
                "ports",
                "revision",
                "firmware",
                "mac",
                "ip",
                "network",
                "led",
            ]
            # keys += [
            #     "host",
            #     "manufacturer",
            #     "model_id",
            #     "description",
            #     "is_eu_model",
            #     "macs",
            #     "hostname",
            # ]
        else:
            keys = ["host"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({vals})"

    def update(self, data: dict) -> None:
        """Update device from given data."""
        self._data.update(data)

    async def refresh(self) -> None:
        """Refresh device data."""
        self._data.update(await self.interface.get_data())

    @property
    def data(self) -> dict[str, Any]:
        """Return device data."""
        return self._data or {}

    @property
    def has_data(self) -> bool:
        """Return whether the device has data."""
        return bool(self._data)

    @property
    def interface(self) -> MPowerInterface:
        """Return device interface."""
        return self._interface

    @property
    def host(self) -> str:
        """Return the device host."""
        return self.interface.host

    @property
    def name(self) -> str:
        """Return the device name."""
        return self.hostname

    @property
    def board_data(self) -> dict[str, Any]:
        """Return the device board data."""
        return self.data.get("board", {})

    @property
    def status_data(self) -> dict[str, Any]:
        """Return the device status data."""
        return self.data.get("status", {})

    @property
    def port_data(self) -> list[dict[str, Any]]:
        """Return port data."""
        return self.data.get("ports", [])

    @property
    def manufacturer(self) -> str:
        """Return the device manufacturer."""
        return "Ubiquiti"

    @property
    def model_id(self) -> str | None:
        """Return the model id."""
        return self.board_data.get("sysid")

    @property
    def ports(self) -> int | None:
        """Return the number of available ports."""
        return self.board_data.get("ports")

    @property
    def description(self) -> str | None:
        """Return the device description as string."""
        ports = self.ports
        if ports:
            if ports == 1:
                return "mFi Power Adapter with Wi-Fi"
            if ports == 3:
                return "3-Port mFi Power Strip with Wi-Fi"
            if ports == 6:
                return "6-Port mFi Power Strip with Ethernet and Wi-Fi"
            if ports == 8:
                return "8-Port mFi Power Strip with Ethernet and Wi-Fi"
        return None

    @property
    def is_eu_model(self) -> bool | None:
        """Return whether this device is a EU model with type F sockets."""
        shortname: str | None = self.board_data.get("shortname")
        if shortname:
            if len(shortname) > 2 and shortname.endswith("E"):
                return True
            elif len(shortname) > 1:
                return False
        return None

    @property
    def model(self) -> str | None:
        """Return the model name."""
        name = self.board_data.get("name")
        eu_tag = " (EU)" if self.is_eu_model else ""
        if name:
            return f"mFi {name}{eu_tag}"
        return None

    @property
    def revision(self) -> str | None:
        """Return the device hardware revision."""
        return self.board_data.get("revision")

    @property
    def mac(self) -> str | None:
        """Return the hardware address from the board."""
        return self.board_data.get("hwaddr")

    @property
    def macs(self) -> list[str] | None:
        """Return hardware addresses from network interfaces."""
        return self.board_data.get("hwaddrs")

    @property
    def firmware(self) -> str | None:
        """Return the device firmware version and build."""
        version = self.status_data.get("firmware_version")
        build = self.status_data.get("firmware_build")
        if build is None:
            return version
        elif version is not None:
            return f"{version} (build {build})"
        return None

    @property
    def hostname(self) -> str:
        """Return the device host name."""
        return self.status_data.get("hostname", self.host)
    
    @property
    def network(self) -> MPowerNetwork | None: 
        """Return the device network interface."""
        return self.status_data.get("iface")

    @property
    def ip(self) -> str | None:
        """Return the device IP address."""
        return self.status_data.get("ipaddr")

    @property
    def led(self) -> MPowerLED | None:
        """Return if the led status."""
        return self.status_data.get("led")

    async def set_led(self, led: MPowerLED, refresh: bool = True) -> None:
        """Set LED state to on/off."""
        await self.interface.set_led(led)
        if refresh:
            await self.refresh()

    async def create_sensor(self, port: int) -> MPowerSensor:
        """Create a single sensor."""
        if not self.has_data:
            await self.refresh()
        return MPowerSensor(self, port)

    async def create_sensors(self) -> list[MPowerSensor]:
        """Create all sensors as list."""
        if not self.has_data:
            await self.refresh()
        return [MPowerSensor(self, i + 1) for i in range(self.ports)]

    async def create_switch(self, port: int) -> MPowerSwitch:
        """Create a single switch."""
        if not self.has_data:
            await self.refresh()
        return MPowerSwitch(self, port)

    async def create_switches(self) -> list[MPowerSwitch]:
        """Create all switches as list."""
        if not self.has_data:
            await self.refresh()
        return [MPowerSwitch(self, i + 1) for i in range(self.ports)]



"""Ubiquiti mFi MPower entities"""

from __future__ import annotations

from . import device
from .interface import MPowerInterface
from .exceptions import MPowerDataError


class MPowerEntity:
    """mFi mPower entity representation."""

    _data: dict
    _device: device.MPowerDevice
    _port: int

    def __init__(
        self,
        device: device.MPowerDevice,
        port: int,
    ) -> None:
        """Initialize the entity."""
        self._data = {}
        self._device = device
        self._port = port
        self.update(device.data)

    def __str__(self):  
        """Represent this entity as string."""
        keys = ["name", "label"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({vals})"

    def update(self, data: dict) -> None:
        """Update entity data from given **device data**."""
        try:
            self._data.update(data.get("ports", [])[self.port - 1])
        except IndexError as exc:
            raise MPowerDataError(
                f"Device data for entity {self.name} is invalid"
            ) from exc

    async def refresh(self) -> None:
        """Refresh entity data from **device data**."""
        await self.device.refresh()
        self.update(self.device.data)

    @property
    def data(self) -> dict:
        """Return entity data."""
        return self._data

    @property
    def has_data(self) -> bool:
        """Return whether the entity has data."""
        return bool(self._data)

    @property
    def device(self) -> device.MPowerDevice:
        """Return the entity device."""
        return self._device

    @property
    def port(self) -> int:
        """Return the entity port."""
        return self._port

    @property
    def host(self) -> str:
        """Return the entity host."""
        return self.device.host

    @property
    def name(self) -> str:
        """Return the device name."""
        return f"{self.device.name}-{self.port}"

    @property
    def interface(self) -> MPowerInterface:
        """Return the device interface."""
        return self.device.interface

    @property
    def label(self) -> str | None:
        """Return the entity label."""
        return self.data.get("config", {}).get("label")

    @property
    def output(self) -> bool | None:
        """Return the current output state."""
        return self.data.get("sensors", {}).get("output")

    @property
    def relay(self) -> bool | None:
        """Return the initial output state which is applied after device boot."""
        return self.data.get("sensors", {}).get("relay")

    @property
    def locked(self) -> bool | None:
        """Return the lock state which prevents switching if enabled."""
        return self.data.get("sensors", {}).get("locked")

    async def set_lock(self, locked: bool, refresh: bool = True) -> None:
        """Set lock state to on/off."""
        await self.interface.set_port_lock(self.port, locked)
        if refresh:
            await self.refresh()

    async def lock(self, refresh: bool = True) -> None:
        """Lock output switch."""
        await self.set_lock(True, refresh=refresh)

    async def unlock(self, refresh: bool = True) -> None:
        """Unlock output switch."""
        await self.set_lock(False, refresh=refresh)

class MPowerSensor(MPowerEntity):
    """mFi mPower sensor representation."""

    def __str__(self):
        """Represent this sensor as string."""
        keys = ["name", "label", "power", "current", "voltage", "powerfactor", "energy"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({vals})"

    @property
    def power(self) -> float | None:
        """Return the output power [W]."""
        return self.data.get("sensors", {}).get("power")

    @property
    def current(self) -> float | None:
        """Return the output current [A]."""
        return self.data.get("sensors", {}).get("current")

    @property
    def voltage(self) -> float | None:
        """Return the output voltage [V]."""
        return self.data.get("sensors", {}).get("voltage")

    @property
    def powerfactor(self) -> float | None:
        """Return the output power factor ("real power" / "apparent power")."""
        return self.data.get("sensors", {}).get("powerfactor")

    @property
    def energy(self) -> float | None:
        """Return the energy since last device boot [Wh]."""
        return self.data.get("sensors", {}).get("energy")


class MPowerSwitch(MPowerEntity):
    """mFi mPower switch representation."""

    def __str__(self):
        """Represent this switch as string."""
        keys = ["name", "label", "output", "relay", "locked"]
        vals = ", ".join([f"{k}={getattr(self, k)}" for k in keys])
        return f"{__class__.__name__}({vals})"

    async def set_output(self, output: bool, refresh: bool = True) -> None:
        """Set output to on/off."""
        await self.interface.set_port_output(self.port, output)
        if refresh:
            await self.refresh()

    async def turn_on(self, refresh: bool = True) -> None:
        """Turn output on."""
        await self.set_output(True, refresh=refresh)

    async def turn_off(self, refresh: bool = True) -> None:
        """Turn output off."""
        await self.set_output(False, refresh=refresh)
    async def toggle(self, refresh: bool = True) -> None:
        """Toggle output."""
        await self.set_output(not self.output, refresh=refresh)



"""Ubiquiti mFi MPower SSH session"""

from __future__ import annotations

import asyncio
from collections.abc import AsyncGenerator
from contextlib import asynccontextmanager

import asyncssh

from .exceptions import MPowerError, MPowerDataError


class MPowerConnectionError(MPowerError):
    """Error related to connections."""


class MPowerAuthenticationError(MPowerError):
    """Error related to authentication."""


class MPowerCommandError(MPowerError):
    """Error related to command execution."""


class MPowerSession:
    """mFi mPower session representation."""

    _host: str
    _username: str
    _password: str
    _conn: asyncssh.SSHClientConnection | None
    _lock: asyncio.Lock

    # NOTE: Ubiquiti mFi mPower Devices with firmware 2.1.11 use Dropbear SSH 0.51 (27 Mar 2008).
    options: dict = {
        "kex_algs": "diffie-hellman-group1-sha1",
        "encryption_algs": "aes128-cbc",
        # https://github.com/ronf/asyncssh/issues/263
        "server_host_key_algs": "ssh-rsa",
        # https://github.com/ronf/asyncssh/issues/132
        "known_hosts": None,
    }

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
    ) -> None:
        """Initialize the session."""
        self._host = host
        self._username = username
        self._password = password
        self._conn = None
        self._lock = asyncio.Lock()

    @property
    def host(self) -> str:
        """Return the host."""
        return self._host

    async def connect(self) -> None:
        """Establish SSH connection."""
        if self._conn:
            self._conn.close()
            await self._conn.wait_closed()
        try:
            self._conn = await asyncssh.connect(
                host=self._host,
                username=self._username,
                password=self._password,
                **self.options,
            )
        except asyncssh.PermissionDenied as exc:
            raise MPowerAuthenticationError(
                f"Login to device {self.host} failed due to wrong SSH credentials"
            ) from exc
        except (OSError, asyncssh.Error) as exc:
            info = f"{type(exc).__name__}({exc})"
            raise MPowerConnectionError(
                f"Connection to device {self.host} failed: {info}"
            ) from exc

    async def close(self) -> None:
        """Close SSH connection."""
        if self._conn:
            self._conn.close()
            await self._conn.wait_closed()
            self._conn = None

    @property
    async def connection(self) -> asyncssh.SSHClientConnection:
        """Return active SSH connection, establishing it if necessary."""
        async with self._lock:
            if self._conn is None or self._conn.is_closed():
                await self.connect()
            return self._conn

    async def run(self, command: str) -> str:
        """Run command over SSH."""
        connection = await self.connection
        process = await connection.run(command)
        status = process.exit_status
        if status != 0:
            raise MPowerCommandError(
                f"Command '{command}' on device {self.host} failed with exit code {status}"
            )
        return process.stdout.strip()
    
    @asynccontextmanager
    async def get(self, command) -> AsyncGenerator[str, None]:
        """Run command over SSH and yield output in context manager."""
        try:
            output = await self.run(command)
            yield output
        except Exception as exc:
            info = f"{type(exc).__name__}({exc})"
            raise MPowerDataError(
                f"Data processing for '{command}' on device {self.host} failed: {info}"
            ) from exc



"""Ubiquiti mFi MPower SSH interface"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import KeysView, ValuesView
from enum import Enum
import json
import re
from typing import Any, Callable

from .session import MPowerSession


class MPowerLED(Enum):
    """mFi mPower LED status representation."""

    OFF = 0
    BLUE = 1
    YELLOW = 2
    BOTH = 3
    ALTERNATE = 4
    LOCKED_OFF = 99  # <turn OFF to unlock>

    def __str__(self):
        return self.name


class MPowerNetwork(Enum):
    """mFi mPower network interface representation."""

    LAN = 0
    WLAN = 1

    @classmethod
    def from_dev(cls, s: str) -> "MPowerNetwork":
        s = s.lower()
        if s.startswith("eth"):
            return cls.LAN
        else:
            return cls.WLAN

    def __str__(self):
        return self.name


class MPowerCat(ABC):
    """mFi mPower cat interface base."""

    _session: MPowerSession

    def __init__(
        self,
        session: MPowerSession,
    ) -> None:
        """Initialize the cat interface."""
        self._session = session

    @property
    def session(self) -> MPowerSession:
        """Return the session."""
        return self._session

    @property
    def dir(self) -> str | None:
        """Return the default directory."""
        return None

    @property
    def file(self) -> str | None:
        """Return the default file."""
        return None

    @property
    def func(self) -> type | Callable:
        """Return the default conversion function."""
        return str

    @property
    @abstractmethod
    def specs(self) -> dict[str, dict[str, Any]]:
        """Return the specs."""
        pass

    def cast(self, value: str, to: type) -> Any:
        if value is None:
            return None
        try:
            return to(value)
        except (ValueError, TypeError):
            return None

    def keys(self) -> KeysView:
        """Return the keys."""
        return self.specs.keys()

    def values(self) -> ValuesView:
        """Return the specs values."""
        return self.specs.values()

    def files(self) -> list[str]:
        """Return the specs files."""
        return [v.get("file", self.file).format(dir=self.dir) for v in self.values()]

    def command(self) -> str:
        """Return the cat command with ASCII Record Separator (0x1E) for file separation."""
        commands = ["cd /tmp"]
        commands += [c for v in self.specs.values() if (c := v.get("command", None))]
        commands += [r"printf '\x1E' > _",  fr"cat {' _ '.join(self.files())}"]
        return " && ".join(commands)

    def convert(self, key: str, values: list[str]) -> Any:
        """Convert cat output values for one spec."""
        func = self.specs[key].get("func", self.func)
        return func(values)

    async def get_data(self) -> dict[str, list[str]]:
        """Run the cat command and parse the output into a data dictionary."""
        keys = list(self.keys())
        async with self.session.get(self.command()) as output:
            outputs = output.split("\x1E")  # Split by separation markers
            groups = {keys[i]: o.splitlines() for i, o in enumerate(outputs)}
            data = {k: self.convert(k, v) for k, v in groups.items()}
            for key in [k for k in keys if self.specs[k].get("unwrap", False)]:
                if isinstance(data[key], dict):  # Flatten (unwrap) only dicts
                    data = {**{kk: vv for kk, vv in data.items() if kk != key}, **data[key]}
            return data
    

class MPowerCatBoard(MPowerCat):
    """mFi mPower cat interface for board information."""

    PORTS = {
        "0xe641": 1,
        "0xe651": 1,
        "0xe671": 1,
        "0xe672": 1,
        "0xe662": 2,
        "0xe643": 3,
        "0xe653": 3,
        "0xe656": 6,
        "0xe648": 8,
    }

    def func_info(self, values: list[str]) -> dict[str, str]:
        """Extract board information."""
        data = {
            match.group(1): match.group(2)
            for value in values
            if (match := re.match(r"^board.(\w+)=(.*)", value))
        }
        data["ports"] = int(self.PORTS.get(data["sysid"].lower(), 0))
        hwaddr = data.pop("hwaddr")
        data["hwaddr"] = ':'.join(hwaddr[i:i+2] for i in range(0, 12, 2))
        return data

    def func_ifconfig(self, values: list[str]) -> dict[str, str]:
        """Extract hardware addresses from ifconfig output."""
        data = {
            "hwaddrs": {
                MPowerNetwork.from_dev(match.group(1)).name.lower(): match.group(2)
                for value in values
                if (match := re.match(r"^([a-zA-Z0-9]+).*?HWaddr ([0-9A-Fa-f:]{17})", value))
            }
        }
        return data
    
    @property
    def specs(self) -> dict[str, dict[str, Any]]:
        """Return the specs."""
        return {
            "info": {"file": "/etc/board.info", "func": self.func_info, "unwrap": True},
            "ifconfig": {
                "file": "ifconfig",
                "command": "ifconfig -a > ifconfig",
                "func": self.func_ifconfig,
                "unwrap": True
            },
        }
    

class MPowerCatStatus(MPowerCat):
    """mFi mPower cat interface for status information."""

    def func(self, values: list[str]) -> str:
        """Unwrap value."""
        return values[0]

    def func_led_status(self, values: list[str]) -> MPowerLED:
        """Unwrap value and convert to LED status enum type."""
        status = self.cast(values[0].split()[0], int)
        return None if status is None else MPowerLED(status)
    
    def func_ip_route(self, values: list[str]) -> dict[str, str]:
        """Extract network interface and IP address from 'ip route' output."""
        data = {
            "iface": match.group(1)
            for value in values
            if (match := re.match(r"^default.* dev (\S+)", value))
        }
        data.update({
            "ipaddr": match.group(1)
            for value in values
            if (match := re.search(fr"dev {data.get('iface')}.*src (\d+\.\d+\.\d+\.\d+)", value))
        })
        data["iface"] = MPowerNetwork.from_dev(data["iface"])
        return data
    
    @property
    def specs(self) -> dict[str, dict[str, Any]]:
        """Return the specs."""
        return {
            "firmware_version": {"file": "/usr/etc/.version"},
            "firmware_build": {"file": "/usr/etc/.build",},
            "led": {"file": "/proc/led/status", "func": self.func_led_status},
            "hostname": {"file": "/proc/sys/kernel/hostname"},
            "ip_route": {
                "file": "ip_route",
                "command": "ip route > ip_route",
                "func": self.func_ip_route,
                "unwrap": True
            },
        }
    

class MPowerCatPort(MPowerCat):
    """mFi mPower cat interface base for ports."""

    _ports: int

    def __init__(
        self,
        session: MPowerSession,
        ports: int,
    ) -> None:
        """Initialize the port cat processor."""
        super().__init__(session)
        self._ports = ports

    @property
    def ports(self) -> int:
        """Return the number of ports."""
        return self._ports


class MPowerCatPortConfig(MPowerCatPort):
    """mFi mPower cat interface for port config data."""

    @property
    def dir(self) -> str:
        return "/etc/persistent/cfg"

    def func_label(self, values: list[str]) -> list[str]:
        """Convert labels."""
        data = {
            int(match.group(1)): match.group(2)
            for value in values
            if (match := re.match(r"port\.(\d+)\.label=(.*)", value))
        }
        label = [data.get(i, None) for i in range(self.ports)]
        return label

    def func_vpower(self, values: list[str]) -> dict[str, bool]:
        """Convert vpower settings."""
        vpower = {}
        for setting in ("enabled", "lock", "relay"):
            data = {
                int(match.group(1))-1: match.group(2)
                for value in values
                if (match := re.match(fr"vpower\.(\d+)\.{setting}=(.*)", value))
            }
            vpower[setting] = [data.get(i, None) for i in range(self.ports)]
        return vpower
    
    @property
    def specs(self) -> dict[str, dict[str, Any]]:
        """Return the specs."""
        return {
            "label": {"file": "{dir}/config_file", "func": self.func_label},
            "vpower": {"file": "{dir}/vpower_cfg", "func": self.func_vpower, "unwrap": True},
        }
    

class MPowerCatPortSensors(MPowerCatPort):
    """mFi mPower cat interface for port sensor data."""

    @property
    def dir(self) -> str:
        return "/proc/power"

    def func_float(self, values: list[str]) -> list[float | None]:
        """Convert float readings."""
        return [self.cast(value, float) for value in values]

    def func_bool(self, values: list[str]) -> list[bool | None]:
        """Convert boolean readings."""
        return [self.cast(self.cast(value, int), bool) for value in values]
    
    @property
    def specs(self) -> dict[str, dict[str, Any]]:
        """Return the specs."""
        return {
            "energy": {"file": "{dir}/energy_sum*", "func": self.func_float},  # energy [Wh]
            "voltage": {"file": "{dir}/v_rms*", "func": self.func_float},  # voltage [V]
            "current": {"file": "{dir}/i_rms*", "func": self.func_float},  # current [A]
            "power": {"file": "{dir}/active_pwr*", "func": self.func_float},  # power [W]
            "powerfactor": {"file": "{dir}/pf*", "func": self.func_float},  # power factor
            "output": {"file": "{dir}/output*", "func": self.func_bool},  # output state
            "enabled": {"file": "{dir}/enabled*", "func": self.func_bool},  # enabled state
            "locked": {"file": "{dir}/lock*", "func": self.func_bool},  # lock state
            "relay": {"file": "{dir}/relay*", "func": self.func_bool},  # relay state
        }


class MPowerInterface:
    """mFi mPower interface representation."""

    _board: dict | None
    _session: MPowerSession

    def __init__(
        self,
        host: str,
        username: str,
        password: str,
    ) -> None:
        """Initialize the interface."""
        self._board = None
        self._session = MPowerSession(host, username, password)

    @property
    def session(self) -> MPowerSession:
        """Return the session."""
        return self._session

    @property
    def host(self) -> str:
        """Return the session host."""
        return self.session.host
        
    async def connect(self) -> None:
        """Establish connection."""
        await self.session.connect()
        
    async def close(self) -> None:
        """Close connection."""
        await self.session.close()
        
    async def run(self, command) -> None:
        """Run command without returning output."""
        await self.session.run(command)

    async def get_board(self) -> dict:
        """Get (stored) board info."""
        if self._board is None:
            self._board = await MPowerCatBoard(self.session).get_data()
        return self._board

    async def get_status_info(self) -> dict:
        """Get device status information."""
        return await MPowerCatStatus(self.session).get_data()

    async def get_ports(self) -> int:
        """Get number of ports from board info."""
        return (await self.get_board())["ports"]

    async def get_port_data_from_cat(self, cat: MPowerCatPort) -> list[dict]:
        """Get port data from template."""
        data = await cat.get_data()
        return [{k: v for k, v in zip(data, values)} for values in zip(*data.values())]

    async def get_port_config_data(self) -> list[dict]:
        """Get port config data."""
        ports = await self.get_ports()
        return await self.get_port_data_from_cat(MPowerCatPortConfig(self.session, ports))

    async def get_port_sensor_data(self) -> list[dict]:
        """Get port sensor data."""
        ports = await self.get_ports()
        return await self.get_port_data_from_cat(MPowerCatPortSensors(self.session, ports))

    async def get_port_data(self) -> list[dict]:
        """Get port data."""
        ports = await self.get_ports()
        config = await self.get_port_config_data()
        sensors = await self.get_port_sensor_data()
        return [{"config": config[i], "sensors": sensors[i]} for i in range(ports)]

    async def get_data(self, debug: bool = False) -> dict:
        """Get all data."""
        data = {
            "board": await self.get_board(),
            "status": await self.get_status_info(),
            "ports": await self.get_port_data(),
        }
        
        if debug:
            print("data", "=", json.dumps(data, indent=2, default=str))

        return data

    async def set_proc(self, proc: str, value: Any) -> None:
        """Set state via file."""
        await self.run(f"echo {value} > /proc/{proc}")

    async def set_led(self, led: MPowerLED) -> None:
        """Set LED state to the given LED value."""
        await self.set_proc("led/status", MPowerLED.OFF.value)
        await self.set_proc("led/status", led.value)

    async def set_port_lock(self, port: int, lock: bool) -> None:
        """Set port lock state to on/off."""
        await self.set_proc(f"power/lock{port}", int(lock))

    async def set_port_output(self, port: int, output: bool) -> None:
        """Set port output state to on/off."""
        await self.set_proc(f"power/output{port}", int(output))